#!/usr/bin/python

from distutils.core import setup

packages = ['MoinMoin', 'MoinMoin.action', 'MoinMoin.formatter']

setup( name = 'MoinMoin2ODF', 
       description = 'MoinMoin OpenDocument Export',
       author = 'Hans-Peter Schaal',
       author_email = 'hp.news@gmx.de',
       version = '0.9', 
       packages = packages )
